import fnmatch
import click
import collections
import mlflow
import json
import time
import os.path

from ast import literal_eval
from tempfile import TemporaryDirectory
from tqdm import tqdm

from aim import Run, Image, Text, Audio

IMAGE_EXTENSIONS = ('jpg', 'bmp', 'jpeg', 'png', 'gif', 'svg')
HTML_EXTENSIONS = ('html',)
TEXT_EXTENSIONS = (
    'txt',
    'log',
    'py',
    'js',
    'yaml',
    'yml',
    'json',
    'csv',
    'tsv',
    'md',
    'rst',
    'jsonnet',
)

# Audio is not handled in mlflow but including here just in case
AUDIO_EXTENSIONS = (
    'flac',
    'mp3',
    'wav',
)


class RunHashCache:
    def __init__(self, repo_path, no_cache=False):
        self._cache_path = os.path.join(repo_path, 'mlflow_logs_cache')
        self._needs_refresh = False

        if no_cache and os.path.exists(self._cache_path):
            os.remove(self._cache_path)

        try:
            with open(self._cache_path) as FS:
                self._cache = json.load(FS)
        except Exception:
            self._cache = {}

    def get(self, run_id):
        return self._cache.get(run_id)

    def __setitem__(self, key: str, val: str):
        if not self._cache.get(key) == val:
            self._cache[key] = val
            self._needs_refresh = True

    def __getitem__(self, key: str):
        return self._cache[key]

    def refresh(self):
        if self._needs_refresh:
            with open(self._cache_path, 'w') as FS:
                json.dump(self._cache, FS)


def get_mlflow_experiments(client, experiment):
    if experiment is None:
        # process all experiments
        experiments = client.search_experiments()
    else:
        try:
            ex = client.get_experiment(experiment)
        except mlflow.exceptions.MlflowException:
            ex = client.get_experiment_by_name(experiment)
        if not ex:
            raise RuntimeError(f'Could not find experiment with id or name "{experiment}"')
        experiments = (ex,)

    return experiments


def get_aim_run(repo_inst, run_id, run_name, experiment_name, run_cache):
    if run_cache.get(run_id):
        aim_run = Run(
            run_hash=run_cache[run_id],
            repo=repo_inst,
            system_tracking_interval=None,
            capture_terminal_logs=False,
            experiment=experiment_name,
        )
    else:
        aim_run = Run(
            repo=repo_inst,
            system_tracking_interval=None,
            capture_terminal_logs=False,
            experiment=experiment_name,
        )
        run_cache[run_id] = aim_run.hash
    aim_run.name = run_name
    return aim_run


def collect_run_params(aim_run, mlflow_run):
    aim_run['mlflow_run_id'] = mlflow_run.info.run_id
    aim_run['mlflow_experiment_id'] = mlflow_run.info.experiment_id
    aim_run.description = mlflow_run.data.tags.get("mlflow.note.content")

    # Collect params & tags
    # MLflow provides "string-ified" params values and we try to revert that
    aim_run['params'] = _map_nested_dicts(_try_parse_str, mlflow_run.data.params)
    aim_run['tags'] = {
        k: v for k, v in mlflow_run.data.tags.items() if not k.startswith('mlflow')
    }


def collect_artifacts(aim_run, mlflow_run, mlflow_client, exclude_artifacts):
    if '*' in exclude_artifacts:
        return

    run_id = mlflow_run.info.run_id

    artifacts_cache_key = '_mlflow_artifacts_cache'
    artifacts_cache = aim_run.meta_run_tree.get(artifacts_cache_key) or []

    __html_warning_issued = False
    with TemporaryDirectory(prefix=f'mlflow_{run_id}_') as temp_path:
        artifact_loc_stack = [None]
        while artifact_loc_stack:
            loc = artifact_loc_stack.pop()
            artifacts = mlflow_client.list_artifacts(run_id, path=loc)

            for file_info in artifacts:
                if file_info.is_dir:
                    artifact_loc_stack.append(file_info.path)
                    continue

                if file_info.path in artifacts_cache:
                    continue
                else:
                    artifacts_cache.append(file_info.path)

                if exclude_artifacts:
                    exclude = False
                    for expr in exclude_artifacts:
                        if fnmatch.fnmatch(file_info.path, expr):
                            exclude = True
                            break
                    if exclude:
                        continue

                downloaded_path = mlflow_client.download_artifacts(run_id, file_info.path, dst_path=temp_path)
                if file_info.path.endswith(HTML_EXTENSIONS):
                    if not __html_warning_issued:
                        click.secho(
                            'Handler for html file types is not yet implemented.', fg='yellow'
                        )
                        __html_warning_issued = True
                    continue
                elif file_info.path.endswith(IMAGE_EXTENSIONS):
                    aim_object = Image
                    kwargs = dict(
                        image=downloaded_path,
                        caption=file_info.path
                    )
                    content_type = 'image'
                elif file_info.path.endswith(TEXT_EXTENSIONS):
                    with open(downloaded_path) as fh:
                        content = fh.read()
                    aim_object = Text
                    kwargs = dict(
                        text=content
                    )
                    content_type = 'text'
                elif file_info.path.endswith(AUDIO_EXTENSIONS):
                    audio_format = os.path.splitext(file_info.path)[1].lstrip('.')
                    aim_object = Audio
                    kwargs = dict(
                        data=downloaded_path,
                        caption=file_info.path,
                        format=audio_format
                    )
                    content_type = 'audio'
                else:
                    click.secho(
                        f'Unresolved or unsupported type for artifact {file_info.path}', fg='yellow'
                    )
                    continue

                try:
                    item = aim_object(**kwargs)
                except Exception as exc:
                    click.echo(
                        f'Could not convert artifact {file_info.path} into aim object - {exc}', err=True
                    )
                    continue
                aim_run.track(item, name=loc or 'root', context={'type': content_type})

            aim_run.meta_run_tree[artifacts_cache_key] = artifacts_cache


def collect_metrics(aim_run, mlflow_run, mlflow_client, timestamp=None):
    for key in mlflow_run.data.metrics.keys():
        metric_history = mlflow_client.get_metric_history(mlflow_run.info.run_id, key)
        if timestamp:
            metric_history = list(filter(lambda m: m.timestamp >= timestamp, metric_history))

        for m in metric_history:
            aim_run.track(m.value, step=m.step, name=m.key)


def convert_existing_logs(repo_inst, tracking_uri, experiment=None, excluded_artifacts=None, no_cache=False):
    client = mlflow.tracking.client.MlflowClient(tracking_uri=tracking_uri)

    experiments = get_mlflow_experiments(client, experiment)
    run_cache = RunHashCache(repo_inst.path, no_cache)
    for ex in tqdm(experiments, desc=f'Parsing mlflow experiments in {tracking_uri}', total=len(experiments)):
        runs = client.search_runs(ex.experiment_id)

        for run in tqdm(runs, desc=f'Parsing mlflow runs for experiment `{ex.name}`', total=len(runs)):
            # get corresponding `aim.Run` object for mlflow run
            aim_run = get_aim_run(repo_inst, run.info.run_id, run.info.run_name, ex.name, run_cache)
            # Collect params and tags
            collect_run_params(aim_run, run)

            # Collect metrics
            collect_metrics(aim_run, run, client)

            # Collect artifacts
            collect_artifacts(aim_run, run, client, excluded_artifacts)

    run_cache.refresh()


def _wait_forever(watcher):
    try:
        while True:
            time.sleep(24 * 60 * 60)  # sleep for a day
    except KeyboardInterrupt:
        watcher.stop()


def _map_nested_dicts(fun, tree):
    if isinstance(tree, collections.abc.Mapping):
        return {k: _map_nested_dicts(fun, subtree) for k, subtree in tree.items()}
    else:
        return fun(tree)


def _try_parse_str(s):
    assert isinstance(s, str), f'Expected a string, got {s} of type {type(s)}'
    try:
        return literal_eval(s.strip())
    except:  # noqa: E722
        return s



import logging
import time

from threading import Thread
from typing import Union, TYPE_CHECKING, Dict

import mlflow.entities
from mlflow import MlflowClient
from aimlflow.utils import (
    get_mlflow_experiments,
    get_aim_run,
    collect_metrics,
    collect_artifacts,
    collect_run_params,
    RunHashCache
)

if TYPE_CHECKING:
    from aim import Repo, Run

logger = logging.getLogger(__name__)


class MLFlowWatcher:
    WATCH_INTERVAL_DEFAULT = 10

    def __init__(self,
                 repo: 'Repo',
                 tracking_uri: str,
                 experiment: str = None,
                 exclude_artifacts: str = None,
                 interval: Union[int, float] = WATCH_INTERVAL_DEFAULT,
                 ):

        self._last_watch_time = time.time()
        self._active_aim_runs_pool: Dict[str, 'Run'] = dict()

        self._watch_interval = interval

        self._client = MlflowClient(tracking_uri)

        self._exclude_artifacts = exclude_artifacts
        self._experiment = experiment
        self._experiments = get_mlflow_experiments(self._client, self._experiment)
        self._repo = repo

        self._th_collector = Thread(target=self._watch, daemon=True)
        self._shutdown = False
        self._started = False

    def start(self):
        if self._started:
            return

        self._started = True
        self._th_collector.start()

    def stop(self):
        if not self._started:
            return

        self._shutdown = True
        self._th_collector.join()

    def _search_experiment(self, experiment_id):
        return next((exp for exp in self._experiments if exp.experiment_id == experiment_id), None)

    def _get_current_active_mlflow_runs(self):
        experiment_ids = [ex.experiment_id for ex in self._experiments]

        active_runs = self._client.search_runs(
            experiment_ids=experiment_ids,
            run_view_type=mlflow.entities.ViewType.ACTIVE_ONLY,
            filter_string='attribute.status="RUNNING"'
        )

        return active_runs

    def _process_single_run(self, aim_run, mlflow_run):
        # Collect params and tags
        collect_run_params(aim_run, mlflow_run)

        # Collect metrics
        collect_metrics(aim_run, mlflow_run, self._client, timestamp=self._last_watch_time)

        # Collect artifacts
        collect_artifacts(aim_run, mlflow_run, self._client, self._exclude_artifacts)

    def _process_runs(self):
        watch_started_time = time.time()

        # refresh experiments list
        self._experiments = get_mlflow_experiments(self._client, self._experiment)

        # process active runs
        active_mlflow_runs = self._get_current_active_mlflow_runs()

        run_cache = RunHashCache(self._repo.path)
        active_mlflow_run_ids = set()

        for mlflow_run in active_mlflow_runs:
            mlflow_run_id = mlflow_run.info.run_id
            active_mlflow_run_ids.add(mlflow_run_id)
            mlflow_experiment = self._search_experiment(mlflow_run.info.experiment_id)
            if self._active_aim_runs_pool.get(mlflow_run_id):
                aim_run = self._active_aim_runs_pool[mlflow_run_id]
            else:
                aim_run = get_aim_run(self._repo,
                                      mlflow_run.info.run_id,
                                      mlflow_run.info.run_name,
                                      mlflow_experiment.name,
                                      run_cache)
                self._active_aim_runs_pool[mlflow_run_id] = aim_run

            self._process_single_run(aim_run, mlflow_run)

        # process closed runs
        all_mlflow_run_ids = set(self._active_aim_runs_pool.keys())
        closed_mlflow_run_ids = all_mlflow_run_ids.difference(active_mlflow_run_ids)

        for mlflow_run_id in closed_mlflow_run_ids:
            # process closed run and remove from pool
            mlflow_run = self._client.get_run(mlflow_run_id)
            aim_run = self._active_aim_runs_pool[mlflow_run_id]
            self._process_single_run(aim_run, mlflow_run)
            aim_run.close()
            del self._active_aim_runs_pool[mlflow_run_id]

        # refresh runs cache and update timer
        self._last_watch_time = watch_started_time
        run_cache.refresh()

    def _watch(self):
        self._process_runs()
        watch_interval_counter = 0
        while True:
            if self._shutdown:
                break

            time.sleep(1)
            watch_interval_counter += 1

            if watch_interval_counter > self._watch_interval:
                self._process_runs()
                watch_interval_counter = 0



import sys
import os
from shutil import rmtree
from setuptools import find_packages, setup, Command

version_file = 'aimlflow/VERSION'

__version__ = None
with open(version_file) as vf:
    __version__ = vf.read().strip()


here = os.path.abspath(os.path.dirname(__file__))

# Package meta-data
NAME = 'aim-mlflow'
DESCRIPTION = 'Aim-MLflow integration'
VERSION = __version__
REQUIRES_PYTHON = '>=3.7.0'

readme_file = 'README.md'
readme_text = open('/'.join((here, readme_file)), encoding="utf-8").read()
LONG_DESCRIPTION = readme_text.strip()

# Get packages
packages = find_packages(exclude=('tests', 'examples'))
version_files = ['../aimlflow/VERSION', ]

package_data = {'aimlflow': version_files}


REQUIRED = [
    'aim',
    'mlflow',
    'tqdm',
]


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(VERSION))
        os.system('git push --tags')

        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    python_requires=REQUIRES_PYTHON,
    install_requires=REQUIRED,
    packages=packages,
    package_data=package_data,
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    cmdclass={
        'upload': UploadCommand
    },
    entry_points={
        'console_scripts': [
            'aimlflow=aimlflow.cli:cli_entry_point',
        ],
    },
)





import click
import os


from click import core
from click import ClickException

from aim.sdk.repo import Repo
from aim.sdk.utils import clean_repo_path

from aimlflow.utils import convert_existing_logs
from aimlflow.watcher import MLFlowWatcher
core._verify_python3_env = lambda: None


@click.group()
def cli_entry_point():
    pass


@cli_entry_point.command(name='sync')
@click.option('--aim-repo', required=False, type=click.Path(exists=True,
                                                            file_okay=False,
                                                            dir_okay=True,
                                                            writable=True))
@click.option('--mlflow-tracking-uri',  required=False, default=None)
@click.option('--experiment', '-e', required=False, default=None)
@click.option('--exclude-artifacts', multiple=True, required=False)
def sync(aim_repo, mlflow_tracking_uri, experiment, exclude_artifacts):

    repo_path = clean_repo_path(aim_repo) or Repo.default_repo_path()
    repo_inst = Repo.from_path(repo_path)

    mlflow_tracking_uri = mlflow_tracking_uri or os.environ.get('MLFLOW_TRACKING_URI')
    if not mlflow_tracking_uri:
        raise ClickException('MLFlow tracking URI must be provided either through ENV or CLI.')

    watcher = MLFlowWatcher(repo_inst, mlflow_tracking_uri, experiment, exclude_artifacts)

    click.echo('Converting existing MLflow logs.')
    convert_existing_logs(repo_inst, mlflow_tracking_uri, experiment, exclude_artifacts)

    click.echo(f'Starting watcher on {mlflow_tracking_uri}.')
    watcher.start()
    from aimlflow.utils import _wait_forever
    _wait_forever(watcher)
